﻿#include "pch.h"
#include "PlatonicSolidsRenderer.h"
#include "Common/DirectXHelper.h"

using namespace PlatonicSolids;

using namespace D2D1;
using namespace DirectX;
using namespace Microsoft::WRL;

PlatonicSolidsRenderer::PlatonicSolidsRenderer(const std::shared_ptr<DX::DeviceResources>& deviceResources) : 
    m_deviceResources(deviceResources),

    // Initialize for 5 platonic solids
    m_figureInfos(std::vector<FigureInfo>(5))
{
    // Tetrahedron
    FigureInfo tetrahedron;
    tetrahedron.faceCount = 4;
    tetrahedron.trianglesPerFace = 1;
    tetrahedron.srcTriangles = 
    {
        Triangle3D { XMFLOAT4(-1,  1, -1, 1), 
                        XMFLOAT4(-1, -1,  1, 1), 
                        XMFLOAT4( 1,  1,  1, 1) },

        Triangle3D { XMFLOAT4( 1, -1, -1, 1), 
                        XMFLOAT4( 1,  1,  1, 1), 
                        XMFLOAT4(-1, -1,  1, 1) },

        Triangle3D { XMFLOAT4( 1,  1,  1, 1), 
                        XMFLOAT4( 1, -1, -1, 1), 
                        XMFLOAT4(-1,  1, -1, 1) },

        Triangle3D { XMFLOAT4(-1, -1,  1, 1), 
                        XMFLOAT4(-1,  1, -1, 1), 
                        XMFLOAT4( 1, -1, -1, 1) }
    };

    tetrahedron.srcTriangles.shrink_to_fit();
    tetrahedron.dstTriangles.resize(tetrahedron.srcTriangles.size());
    tetrahedron.color = ColorF(ColorF::Magenta);
    tetrahedron.renderInfo.resize(tetrahedron.faceCount);
    m_figureInfos.at(0) = tetrahedron;

    // Octahedron
    FigureInfo octahedron;
    octahedron.faceCount = 8;
    octahedron.trianglesPerFace = 1;
    octahedron.srcTriangles =
    {
        Triangle3D{ XMFLOAT4(0, 1, 0, 1), 
                    XMFLOAT4(0, 0, 1, 1), 
                    XMFLOAT4(1, 0, 0, 1) },

        Triangle3D{ XMFLOAT4(0, 1, 0, 1), 
                    XMFLOAT4(-1, 0, 0, 1), 
                    XMFLOAT4(0, 0, 1, 1) },

        Triangle3D{ XMFLOAT4(0, -1, 0, 1), 
                    XMFLOAT4(0, 0, 1, 1), 
                    XMFLOAT4(-1, 0, 0, 1) },

        Triangle3D{ XMFLOAT4(0, -1, 0, 1), 
                    XMFLOAT4(1, 0, 0, 1), 
                    XMFLOAT4(0, 0, 1, 1) },

        Triangle3D{ XMFLOAT4(0, -1, 0, 1), 
                    XMFLOAT4(0, 0, -1, 1), 
                    XMFLOAT4(1, 0, 0, 1) },

        Triangle3D{ XMFLOAT4(0, -1, 0, 1), 
                    XMFLOAT4(-1, 0, 0, 1), 
                    XMFLOAT4(0, 0, -1, 1) },

        Triangle3D{ XMFLOAT4(0, 1, 0, 1), 
                    XMFLOAT4(0, 0, -1, 1), 
                    XMFLOAT4(-1, 0, 0, 1) },

        Triangle3D{ XMFLOAT4(0, 1, 0, 1), 
                    XMFLOAT4(1, 0, 0, 1), 
                    XMFLOAT4(0, 0, -1, 1) }
    };

    octahedron.srcTriangles.shrink_to_fit();
    octahedron.dstTriangles.resize(octahedron.srcTriangles.size());
    octahedron.color = ColorF(ColorF::Cyan);
    octahedron.renderInfo.resize(octahedron.faceCount);
    m_figureInfos.at(1) = octahedron;

    // Icosahedron

    // Golden mean, approximately 1.618
    static const float G = (1 + std::sqrt(5.0f)) / 2;   

    FigureInfo icosahedron;
    icosahedron.faceCount = 20;
    icosahedron.trianglesPerFace = 1;
    icosahedron.srcTriangles =
    {
        Triangle3D { XMFLOAT4(0, G, 1, 1), XMFLOAT4(-1, 0, G, 1), XMFLOAT4(1, 0, G, 1) },
        Triangle3D { XMFLOAT4(-1, 0, G, 1), XMFLOAT4(0, -G, 1, 1), XMFLOAT4(1, 0, G, 1) },
        Triangle3D { XMFLOAT4(0, G, 1, 1), XMFLOAT4(1, 0, G, 1), XMFLOAT4(G, 1, 0, 1) },
        Triangle3D { XMFLOAT4(1, 0, G, 1), XMFLOAT4(G, -1, 0, 1), XMFLOAT4(G, 1, 0, 1) },
        Triangle3D { XMFLOAT4(0, G, 1, 1), XMFLOAT4(G, 1, 0, 1), XMFLOAT4(0, G, -1, 1) },
        Triangle3D { XMFLOAT4(G, 1, 0, 1), XMFLOAT4(1, 0, -G, 1), XMFLOAT4(0, G, -1, 1) },
        Triangle3D { XMFLOAT4(0, G, 1, 1), XMFLOAT4(0, G, -1, 1), XMFLOAT4(-G, 1, 0, 1) },
        Triangle3D { XMFLOAT4(0, G, -1, 1), XMFLOAT4(-1, 0, -G, 1), XMFLOAT4(-G, 1, 0, 1) },
        Triangle3D { XMFLOAT4(0, G, 1, 1), XMFLOAT4(-G, 1, 0, 1), XMFLOAT4(-1, 0, G, 1) },
        Triangle3D { XMFLOAT4(-G, 1, 0, 1), XMFLOAT4(-G, -1, 0, 1), XMFLOAT4(-1, 0, G, 1) },
        Triangle3D { XMFLOAT4(1, 0, G, 1), XMFLOAT4(0, -G, 1, 1), XMFLOAT4(G, -1, 0, 1) },
        Triangle3D { XMFLOAT4(0, -G, 1, 1), XMFLOAT4(0, -G, -1, 1), XMFLOAT4(G, -1, 0, 1) },
        Triangle3D { XMFLOAT4(G, 1, 0, 1), XMFLOAT4(G, -1, 0, 1), XMFLOAT4(1, 0, -G, 1) },
        Triangle3D { XMFLOAT4(G, -1, 0, 1), XMFLOAT4(0, -G, -1, 1), XMFLOAT4(1, 0, -G, 1) },
        Triangle3D { XMFLOAT4(0, G, -1, 1), XMFLOAT4(1, 0, -G, 1), XMFLOAT4(-1, 0, -G, 1) },
        Triangle3D { XMFLOAT4(1, 0, -G, 1), XMFLOAT4(0, -G, -1, 1), XMFLOAT4(-1, 0, -G, 1) },
        Triangle3D { XMFLOAT4(-G, 1, 0, 1), XMFLOAT4(-1, 0, -G, 1), XMFLOAT4(-G, -1, 0, 1) },
        Triangle3D { XMFLOAT4(-1, 0, -G, 1), XMFLOAT4(0, -G, -1, 1), XMFLOAT4(-G, -1, 0, 1) },
        Triangle3D { XMFLOAT4(-1, 0, G, 1), XMFLOAT4(-G, -1, 0, 1), XMFLOAT4(0, -G, 1, 1) },
        Triangle3D { XMFLOAT4(-G, -1, 0, 1), XMFLOAT4(0, -G, -1, 1), XMFLOAT4(0, -G, 1, 1) }
    };

    icosahedron.srcTriangles.shrink_to_fit();
    icosahedron.dstTriangles.resize(icosahedron.srcTriangles.size());
    icosahedron.color = ColorF(ColorF::Yellow);
    icosahedron.renderInfo.resize(icosahedron.faceCount);
    m_figureInfos.at(4) = icosahedron;

    // Cube
    PolyhedronDefinition cubeDef = 
    {
        {
            // front
            { { XMFLOAT3(-1,  1, 1), XMFLOAT3(-1, -1, 1),
                XMFLOAT3( 1, -1, 1), XMFLOAT3( 1,  1, 1) } },
            // back
            { { XMFLOAT3( 1,  1, -1), XMFLOAT3( 1, -1, -1),
                XMFLOAT3(-1, -1, -1), XMFLOAT3(-1,  1, -1) } },
            // left
            { { XMFLOAT3(-1,  1, -1), XMFLOAT3(-1, -1, -1),
                XMFLOAT3(-1, -1,  1), XMFLOAT3(-1,  1,  1) } },
            // right
            { { XMFLOAT3(1,  1,  1), XMFLOAT3(1, -1,  1),
                XMFLOAT3(1, -1, -1), XMFLOAT3(1,  1, -1) } },
            // bottom
            { { XMFLOAT3(-1, 1, -1), XMFLOAT3(-1, 1,  1),
                XMFLOAT3( 1, 1,  1), XMFLOAT3( 1, 1, -1) } },
            // top
            { { XMFLOAT3(-1, -1,  1), XMFLOAT3(-1, -1, -1),
                XMFLOAT3( 1, -1, -1), XMFLOAT3( 1, -1,  1) } }
        }
    };

    FigureInfo cube;
    cube.faceCount = (int) cubeDef.faces.size();
    cube.trianglesPerFace = (int) cubeDef.faces.at(0).vertices.size() - 2;
    cube.color = ColorF(ColorF::Red);
    cube.renderInfo.resize(cube.faceCount);

    GenerateTriangles(cubeDef, cube.srcTriangles);
    cube.dstTriangles.resize(cube.srcTriangles.size());
    m_figureInfos.at(2) = cube;

    // Dodecahedron
    static const float H = 1 / G;       // approximately 0.618

    PolyhedronDefinition dodecahedronDef = 
    {
        {
            { { XMFLOAT3(1, -1, -1), XMFLOAT3(G, 0, -H), XMFLOAT3(G, 0, H),
                                     XMFLOAT3(1, -1, 1), XMFLOAT3(H, -G, 0) } },
            { { XMFLOAT3(-1, -1, 1), XMFLOAT3(-G, 0, H), XMFLOAT3(-G, 0, -H), 
                                     XMFLOAT3(-1, -1, -1), XMFLOAT3(-H, -G, 0) } },
            { { XMFLOAT3(-1, 1, -1), XMFLOAT3(-G, 0, -H), XMFLOAT3(-G, 0, H), 
                                     XMFLOAT3(-1, 1, 1), XMFLOAT3(-H, G, 0) } },
            { { XMFLOAT3(1, 1, 1), XMFLOAT3(G, 0, H), XMFLOAT3(G, 0, -H), 
                                   XMFLOAT3(1, 1, -1), XMFLOAT3(H, G, 0) } },
            { { XMFLOAT3(-1, 1, -1), XMFLOAT3(0, H, -G), XMFLOAT3(0, -H, -G), 
                                     XMFLOAT3(-1, -1, -1), XMFLOAT3(-G, 0, -H) } },
            { { XMFLOAT3(-1, -1, 1), XMFLOAT3(0, -H, G), XMFLOAT3(0, H, G), 
                                     XMFLOAT3(-1, 1, 1), XMFLOAT3(-G, 0, H) } },
            { { XMFLOAT3(1, -1, -1), XMFLOAT3(0, -H, -G), XMFLOAT3(0, H, -G), 
                                     XMFLOAT3(1, 1, -1), XMFLOAT3(G, 0, -H) } },
            { { XMFLOAT3(1, 1, 1), XMFLOAT3(0, H, G), XMFLOAT3(0, -H, G), 
                                   XMFLOAT3(1, -1, 1), XMFLOAT3(G, 0, H) } },
            { { XMFLOAT3(1, -1, -1), XMFLOAT3(H, -G, 0), XMFLOAT3(-H, -G, 0), 
                                     XMFLOAT3(-1, -1, -1), XMFLOAT3(0, -H, -G) } },
            { { XMFLOAT3(-1, 1, -1), XMFLOAT3(-H, G, 0), XMFLOAT3(H, G, 0), 
                                     XMFLOAT3(1, 1, -1), XMFLOAT3(0, H, -G) } },
            { { XMFLOAT3(-1, -1, 1), XMFLOAT3(-H, -G, 0), XMFLOAT3(H, -G, 0), 
                                     XMFLOAT3(1, -1, 1), XMFLOAT3(0, -H, G) } },
            { { XMFLOAT3(1, 1, 1), XMFLOAT3(H, G, 0), XMFLOAT3(-H, G, 0), 
                                   XMFLOAT3(-1, 1, 1), XMFLOAT3(0, H, G) } }
        }
    };

    FigureInfo dodecahedron;
    dodecahedron.faceCount = (int) dodecahedronDef.faces.size();
    dodecahedron.trianglesPerFace = (int) dodecahedronDef.faces.at(0).vertices.size() - 2;
    dodecahedron.color = ColorF(ColorF::Blue);
    dodecahedron.renderInfo.resize(dodecahedron.faceCount);

    GenerateTriangles(dodecahedronDef, dodecahedron.srcTriangles);
    dodecahedron.dstTriangles.resize(dodecahedron.srcTriangles.size());
    m_figureInfos.at(3) = dodecahedron;

    DX::ThrowIfFailed(
        m_deviceResources->GetD2DFactory()->CreateDrawingStateBlock(&m_stateBlock)
        );

    CreateDeviceDependentResources();
}

void PlatonicSolidsRenderer::GenerateTriangles(PolyhedronDefinition& polyDef,
                                               std::vector<Triangle3D>& triangles)
{
    for (FaceDefinition& faceDef : polyDef.faces)
    {
        for (unsigned int vertex = 0; vertex < faceDef.vertices.size() - 2; vertex++)
        {
            Triangle3D triangle3D;

            XMFLOAT3 point1 = faceDef.vertices.at(0);
            triangle3D.point1 = XMFLOAT4(point1.x, point1.y, point1.z, 1);

            XMFLOAT3 point2 = faceDef.vertices.at(vertex + 1);
            triangle3D.point2 = XMFLOAT4(point2.x, point2.y, point2.z, 1);

            XMFLOAT3 point3 = faceDef.vertices.at(vertex + 2);
            triangle3D.point3 = XMFLOAT4(point3.x, point3.y, point3.z, 1);

            triangles.push_back(triangle3D);
        }
    }

    triangles.shrink_to_fit();
}

void PlatonicSolidsRenderer::CreateDeviceDependentResources()
{
    // Create solid color brushes
    for (FigureInfo& figureInfo : m_figureInfos)
    {
        for (RenderInfo& renderInfo : figureInfo.renderInfo)
        {
            m_deviceResources->GetD2DDeviceContext()->CreateSolidColorBrush(ColorF(0), &renderInfo.brush);
        }
    }
}

void PlatonicSolidsRenderer::ReleaseDeviceDependentResources()
{
    // Destroy solid color brushes
    for (FigureInfo& figureInfo : m_figureInfos)
    {
        for (RenderInfo& renderInfo : figureInfo.renderInfo)
        {
            renderInfo.brush.Reset();
        }
    }
}

void PlatonicSolidsRenderer::Update(DX::StepTimer const& timer)
{
    static const float cycleDuration = 30;
    float cycleTime = fmod((float) timer.GetTotalSeconds(), cycleDuration);

    // Calculate rotations
    static const float animaBeginTime = 2.5f;
    static float xAngle = 0, yAngle = 0, zAngle = 0;
    float animaTime = (float)timer.GetElapsedSeconds();

    if (cycleTime > animaBeginTime)
    {
        // Rotation around X axis
        float period = 7;
        xAngle += XM_2PI * fmod(animaTime, period) / period;

        // Rotation around Y axis
        period = 11;
        yAngle += XM_2PI * fmod(animaTime, period) / period;

        // Rotation around Z axis
        period = 13;
        zAngle += XM_2PI * fmod(animaTime, period) / period;
    }

    // Find scaling based on window size
    Windows::Foundation::Size logicalSize = 
                            m_deviceResources->GetLogicalSize();
    float scale = min(logicalSize.Width, logicalSize.Height) / 10.0f;

    // Calculate total matrix
    XMMATRIX matrix = XMMatrixScaling(scale, scale, scale) * 
                        XMMatrixRotationX(xAngle) * 
                        XMMatrixRotationY(yAngle) * 
                        XMMatrixRotationZ(zAngle);

    // Transform source triangles to destination triangles
    for (FigureInfo& figureInfo : m_figureInfos)
    {
        XMVector4TransformStream(
            (XMFLOAT4 *) figureInfo.dstTriangles.data(),
            sizeof(XMFLOAT4),
            (XMFLOAT4 *) figureInfo.srcTriangles.data(),
            sizeof(XMFLOAT4),
            3 * figureInfo.srcTriangles.size(),
            matrix);
    }

    // Calculate animated intensity of directional light
    static const float lightBeginTime = 5;
    float lightIntensity = 0;

    if (cycleTime > lightBeginTime)
    {
        float relativeTime = (cycleTime - lightBeginTime) / 
                             (cycleDuration - lightBeginTime);

        static const float rampUp = 0.5f;
        static const float rampDown = 0.75f;

        if (relativeTime < rampUp)
        {
            lightIntensity = relativeTime / rampUp;
        }
        else if (relativeTime < rampDown)
        {
            lightIntensity = 1;
        }
        else
        {
            lightIntensity = 1 - (relativeTime - rampDown) / 
                                 (1 - rampDown);
        }
    }

    for (FigureInfo& figureInfo : m_figureInfos)
    {
        // Loop through the faces
        for (int face = 0; face < figureInfo.faceCount; face++)
        {
            // Create a mesh for each face
            RenderInfo& renderInfo = figureInfo.renderInfo.at(face);
            
            DX::ThrowIfFailed(
                m_deviceResources->GetD2DDeviceContext()->CreateMesh(&renderInfo.mesh)
                );

            ComPtr<ID2D1TessellationSink> tessellationSink;

            DX::ThrowIfFailed(
                renderInfo.mesh->Open(&tessellationSink)
                );

            // Loop through the triangles that comprise the face
            for (int tri = 0; tri < figureInfo.trianglesPerFace; tri++)
            {
                // Collapse each triangle in the face from 3D to 2D
                Triangle3D& triangle3D = figureInfo.dstTriangles.at(figureInfo.trianglesPerFace * face + tri);

                D2D1_TRIANGLE triangle;
                triangle.point1 = Point2F(triangle3D.point1.x, triangle3D.point1.y);
                triangle.point2 = Point2F(triangle3D.point2.x, triangle3D.point2.y);
                triangle.point3 = Point2F(triangle3D.point3.x, triangle3D.point3.y);

                // Add the triangle to the mesh
                tessellationSink->AddTriangles(&triangle, 1);
            }

            tessellationSink->Close();

            // Get one triangle from the face
            Triangle3D& triangle3D = figureInfo.dstTriangles.at(figureInfo.trianglesPerFace * face);

            // Load the 3 points into XMVECTOR registers
            XMVECTOR point1 = XMLoadFloat4(&triangle3D.point1);
            XMVECTOR point2 = XMLoadFloat4(&triangle3D.point2);
            XMVECTOR point3 = XMLoadFloat4(&triangle3D.point3);

            // Find vectors from point1
            XMVECTOR v1 = XMVectorSubtract(point2, point1);
            XMVECTOR v2 = XMVectorSubtract(point3, point1);

            // Find the normal from the cross product
            XMVECTOR normal = XMVector3Cross(v1, v2);
            normal = XMVector3Normalize(normal);

            // Define a direction for direction light
            XMVECTOR lightVector = XMVectorSet(2, 3, -1, 0);
            lightVector = XMVector3Normalize(lightVector);

            // Get base color and initialize render color to transparent
            D2D1_COLOR_F baseColor = figureInfo.color;
            ColorF renderColor = ColorF(0, 0, 0, 0);

            // Check if the face is facing towards the viewer
            if (XMVectorGetZ(normal) > 0)
            {
                // Find the dot product of the normal and the light
                XMVECTOR dot = XMVector3Dot(normal, -lightVector);

                // Add ambient light (0.5) and directional light
                float totalLight = 0.5f + 
                    lightIntensity * 0.5f * XMVectorGetX(dot);

                renderColor = ColorF(totalLight * baseColor.r, 
                                     totalLight * baseColor.g, 
                                     totalLight * baseColor.b);
            }
            
            // Set the rendering brush color (could be transparent)
            renderInfo.brush->SetColor(renderColor);
        }
    }
}

// Renders a frame to the screen.
void PlatonicSolidsRenderer::Render()
{
    ID2D1DeviceContext* context = m_deviceResources->GetD2DDeviceContext();
    Windows::Foundation::Size logicalSize = m_deviceResources->GetLogicalSize();

    context->SaveDrawingState(m_stateBlock.Get());
    context->BeginDraw();
    context->Clear(ColorF(ColorF::Black));

    context->SetAntialiasMode(D2D1_ANTIALIAS_MODE_ALIASED);

    // Loop through the FigureInfo's
    for (size_t index = 0; index < m_figureInfos.size(); index++)
    {
        FigureInfo& figureInfo = m_figureInfos.at(index);
    
        // Calculate screen location based on index in collection
        float x = (index == 4) ? 0.5f : 0.25f + 0.5f * (index % 2);
        float y = (index == 4) ? 0.5f : 0.25f + 0.5f * (index / 2);

        D2D1_MATRIX_3X2_F screenTranslation = 
            Matrix3x2F::Translation(x * logicalSize.Width, 
                                    y * logicalSize.Height);

        context->SetTransform(screenTranslation * 
                    m_deviceResources->GetOrientationTransform2D());

        // Loop through meshes to render them
        for (RenderInfo& renderInfo : figureInfo.renderInfo)
        {
            context->FillMesh(renderInfo.mesh.Get(), renderInfo.brush.Get());
        }
    }

    // Ignore D2DERR_RECREATE_TARGET here. This error indicates that the device
    // is lost. It will be handled during the next call to Present.
    HRESULT hr = context->EndDraw();
    if (hr != D2DERR_RECREATE_TARGET)
    {
        DX::ThrowIfFailed(hr);
    }

    context->RestoreDrawingState(m_stateBlock.Get());
}

